#!/usr/bin/python2.5
# -*- coding: utf-8 -*-
import sys
sys.path = ["../.."]+sys.path   # for find MY following google import
from google.appengine.ext import db
from google.appengine.ext.db import polymodel
from google.appengine.api.users import User
import re
import datetime
import xml.etree.ElementTree as ET
import base64
"""
=======================================================================================
AEX - "App Engine eXchange" ©2009 manatlan
=======================================================================================
importe/exporte features for GAE - manatlan@gmail.com

features :
- import/export all gae properties
- import/export all kind of models (model, expando, polymodel)
- serialization thru XML (using the to_xml() instance method)

known issues / future :
- don't know how parent/group should/could work
- APP_ID or AUTH_DOMAIN is totally ignored ;-)
- no transaction during import (but should be easily done)
- Should work with Expando/PolyModel (but not fully tested)

V 0.1 - 2009/04/09 initial release
=======================================================================================

"""


__all__ = ["exporte","importe"]

def strptime(dd):
    g=re.match(r"(\d\d\d\d)\-(\d\d)\-(\d\d) (\d\d):(\d\d):(\d\d)\.?(\d+)?",dd)
    return datetime.datetime(*[int(i) for i in g.groups(0)])


def exporte(*liste):
    """ yield a xml output (utf8)
       'liste' is a list of objects, and object can be a *query* or a *model-like*
    """
    assert len(liste)>0
    yield """<?xml version="1.0" encoding="UTF-8"?>\n"""
    yield """<db xmlns:gd='gdata' xmlns:georss='rss'>\n"""  # TODO: namespace for gd/georss ?
    for object in liste:
        if type(object) == db.Query:
            # it's a query, so we have the entities
            entities = object
        else:
            # it's not a query, it should be a Model-like, so we query all
            entities = object.all()

        if entities.count()>0:
            modelName=entities[0].__class__.__name__
            yield """<%s>\n""" % modelName
            for entity in entities:
                yield entity.to_xml().encode("utf_8") + "\n"
            yield """</%s>\n""" % modelName
    yield """</db>\n"""



def _get_element(real_property,e,keys):
    """ return the value of node property 'e', 'real_property' should be set for
    models (not for expando), and 'keys' is a dict to store reference"""
    name=e.get("name")
    type=e.get("type")
    value=e.text

    if type=="key":
        return keys[value]
    elif type=="atom:category":
        return db.Category(e[0].get('label'))
    elif type=="gd:rating":
        return db.Rating(e[0].get('value'))
    elif type=="atom:link":
        return db.Link(e[0].get('href'))
    elif type=="gd:email":
        return db.Email(e[0].get('address'))
    elif type=="gd:im":
        return db.IM(e[0].get('protocol'),e[0].get('address'))
    elif type=="gd:phonenumber":
        return db.PhoneNumber(e[0].text)
    elif type=="gd:postaladdress":
        return db.PostalAddress(e[0].text)
    elif type=="georss:point":
        return db.GeoPt(*e[0].text.split(" "))
    elif type=="user":
        return User(value)
    elif type=="int":
        return int(value)
    elif type=="float":
        return float(value)
    elif type=="bool":
        return (value=="True")
    elif type=="blob":
        return db.Blob(base64.urlsafe_b64decode(value))
    elif type=="bytestring":
        return value.decode("base64")
    elif type=="gd:when":
        dt=strptime(value)
        if real_property == datetime.datetime:
            return dt
        elif real_property == datetime.date:
            return dt.date()
        elif real_property == datetime.time:
            return dt.time()
        else:
            return dt   # for expando, a "gd:when" is a datetime
    else:
        return value


def importe(models,file=None,buffer=None,delete=False):
    if type(models)!=list:
        models=[models,]

    if buffer:
        root=ET.fromstring(buffer)
    elif file:
        root=ET.parse(file).getroot()
    else:
        raise "error"

    table={}
    for model in models:
        table[model.__name__] = model

    if delete:
        for model in models:
            for entity in model.all():
                entity.delete()

    keys={} # dict to store ekey to match reference

    for xmodel in root:
        if xmodel.tag in table:
            model = table[xmodel.tag]
            for entity in xmodel:
                kind= entity.get("kind")

                key= entity.get("key")  # real key

                vals={}
                for e in entity:
                    if e.tag=="key":
                        ekey = e.text   #extended key (always present)
                    else:
                        name=e.get("name")

                        if name in model.properties():
                            real_property = model.properties()[name].data_type
                        else: # an expando property ?
                            real_property=None

                        value = _get_element(real_property,e,keys)

                        if real_property is not None:
                            #real model property
                            if real_property == list:
                                if value is None: value=u""

                                if name in vals:
                                    vals[name].append(value)
                                else:
                                    vals[name]=[value]
                            else:
                                vals[name]=value
                        else:
                            # an expando property
                            if name in vals:
                                # it's a list
                                if value is None: value=u""

                                if type(vals[name])!=list:
                                    old = vals[name]
                                    if old is None: old=u""
                                    vals[name] = [old,]

                                vals[name].append(value)
                            else:
                                vals[name]=value


                if "class" in vals: del vals["class"]   #remove class for polymodel

                #~ print vals
                kn=db.Key(key).name()
                if kn:
                    u=model(key_name=kn,**vals)
                else:
                    u=model(**vals)
                u.put()

                #store all ekey reference to entities, to be able to make reference later
                keys[ekey] = u



def _compare(e1,e2):
    assert e1.__class__ == e2.__class__
    model = e1.__class__

    # test static properties()
    for prop,property in model.properties().items():
        v1=getattr(e1,prop)
        v2=getattr(e2,prop)

        if type(property) == db.ReferenceProperty:
            if not (v1 == v2 == None):
                assert _compare(v1,v2)  # can't compare key_name, so compare the content
                #print "*",prop,"REFERENCE OK"
        else:
            #print "*",prop,property.data_type,(v1,v2)
            assert v1==v2, "name:"+prop+"/"+str((v1,v2))


    # test dynamic properties (expando)
    for prop in list(set(e1.dynamic_properties()) | set(e2.dynamic_properties())):
        v1=getattr(e1,prop)
        v2=getattr(e2,prop)

        #~ print "*",prop,type(v1),type(v2),(v1,v2)
        assert v1==v2, "name:"+prop+"/"+str((v1,v2))

    return True

def test_properties():
    class MyModel(db.Model):
        nom = db.StringProperty()
        date  = db.DateTimeProperty()
        date1  = db.DateProperty()
        date2  = db.TimeProperty()
        num=db.IntegerProperty()
        bool=db.BooleanProperty()
        float=db.FloatProperty()
        blob=db.BlobProperty()
        text=db.TextProperty()
        byte=db.ByteStringProperty()
        catg=db.CategoryProperty()
        lists=db.ListProperty(str)
        listi=db.ListProperty(int)
        listsl=db.StringListProperty()
        listc=db.ListProperty(db.Category)
        link=db.LinkProperty()
        email=db.EmailProperty()
        rate=db.RatingProperty()
        user=db.UserProperty()
        phone = db.PhoneNumberProperty()
        addr = db.PostalAddressProperty()
        point = db.GeoPtProperty()
        point2 = db.GeoPtProperty()
        im = db.IMProperty()



    #~ #-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
    # CREATE
    #~ #-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
    u=MyModel()
    u.nom=u"joù"
    u.date=datetime.datetime.now()
    u.date1=datetime.datetime.now().date()
    u.date2=datetime.datetime.now().time()
    u.num=1525
    u.bool=False
    u.float=2.4
    u.blob = db.Blob( "nimp" )
    u.text=u"kâo"
    u.byte="1212"
    u.catg="111" #db.Category("111")
    u.lists=["","v1","v2",u"vé3"]
    u.listi=[0,1,2,3]
    u.listsl=["","v1","v2",u"và3"]
    u.listc=[db.Category(u"ké")]
    u.link="http://gogole.fr"
    u.email="koko@gogole.fr"
    u.rate=87 #db.Rating(87)
    u.user = User("jo@jo.com")
    u.phone = "121213231"
    u.addr = "3, rue de paris"
    u.point=db.GeoPt(12.0,13.0)
    u.point2=db.GeoPt("13.0,15.9")
    u.im=db.IM("http://a","b")
    u.put()
    #~ #-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=


    buf="\n".join( exporte(MyModel) )
    #open("dbx.xml","w+").writelines( exporte(MyModel) )

    importe( MyModel, buffer=buf)
    #importe( [MyModel,],file="dbx.xml")


    ll=MyModel.all().fetch(100)
    assert len(ll)==2
    _compare(ll[0],ll[1])


    # we clean the db, to test an empty entity
    for i in MyModel.all():
        i.delete()

    u=MyModel()
    u.put()

    buf="\n".join( exporte(MyModel) )
    importe( MyModel, buffer=buf)
    ll=MyModel.all().fetch(100)
    assert len(ll)==2
    _compare(ll[0],ll[1])



def test_reference():

    class P1(db.Model):
        name = db.StringProperty()

    class P2(db.Model):
        name = db.StringProperty()
        pere = db.ReferenceProperty( P1 )

    p1=P1()
    p1.name="I am p1"
    p1.put()

    p2=P2()
    p2.name=u"I am p2"
    p2.pere = p1
    p2.put()

    buf="\n".join( exporte( P1,P2 ) )   # order is IMPORTANT (b'coz reference)

    importe( [P2,P1],buffer=buf)        # order is not important

    l1=P1.all().fetch(100)
    assert len(l1)==2
    _compare(l1[0],l1[1])

    l2=P2.all().fetch(100)
    assert len(l2)==2
    _compare(l2[0],l2[1])




def test_selfreference():

    class MeOnlyMe(db.Model):
        name = db.StringProperty()
        pere = db.SelfReferenceProperty()

    p1=MeOnlyMe()
    p1.name="I am me"
    p1.put()

    p2=MeOnlyMe()
    p2.name="I am me too"
    p2.pere=p1
    p2.put()

    buf="\n".join( exporte(MeOnlyMe ) )

    importe( MeOnlyMe,buffer=buf)

    ll=MeOnlyMe.all().fetch(100)
    assert len(ll)==4
    _compare(ll[0],ll[2])
    _compare(ll[1],ll[3])


def test_delete_on_import():

    class Obj(db.Model):
        name = db.StringProperty()

    val=u"çàè"

    o=Obj()
    o.name=val
    o.put()

    buf="".join( exporte( Obj ) )

    importe( Obj, buffer=buf)   # no delete
    ll=Obj.all().fetch(100)     # so we will have 2 entities
    assert len(ll)==2
    _compare(ll[0],ll[1])

    importe( Obj, buffer=buf, delete=True) # delete before import
    ll=Obj.all().fetch(100)                #so, we will have 1 entity
    assert len(ll)==1
    assert ll[0].name==val


def test_expando():

    class Oex(db.Expando):
        name = db.StringProperty()

    o=Oex()
    o.name=u"àç"
    o.age=12
    o.tune=12.50
    o.date=datetime.datetime.now()
    o.nimp = None                   # will not be stored
    o.list=["",]+list(" abc")
    o.put()

    buf="".join( exporte( Oex ) )

    importe( Oex, buffer=buf)   # no delete
    ll=Oex.all().fetch(100)     # so we will have 2 entities
    assert len(ll)==2
    _compare(ll[0],ll[1])


def test_poly():
    class Poli(polymodel.PolyModel):
        name = db.StringProperty()

    class PoliMe(Poli):
        name2 = db.StringProperty()

    # test the Herited class : PoliMe
    p=PoliMe()
    p.name="name1"
    p.name2="name2"
    p.put()

    assert p._class == ['Poli', 'PoliMe']

    buf="".join( exporte( PoliMe ) )

    importe( PoliMe, buffer=buf)   # no delete

    ll=PoliMe.all().fetch(100)     # so we will have 2 entities
    assert len(ll)==2
    _compare(ll[0],ll[1])
    assert ll[1]._class == ll[0]._class

    ll=Poli.all().fetch(100)     # so we will have 2 entities
    assert len(ll)==2
    _compare(ll[0],ll[1])
    assert ll[1]._class == ll[0]._class


    for i in PoliMe.all():
        i.delete()

    assert Poli.all().count()==0
    assert PoliMe.all().count()==0

    # test the poly class : Poli
    p=Poli()
    p.name="name1"
    p.put()

    buf="".join( exporte( Poli ) )

    importe( Poli, buffer=buf)   # no delete

    ll=Poli.all().fetch(100)     # so we will have 2 entities
    assert len(ll)==2
    _compare(ll[0],ll[1])

    ll=PoliMe.all().fetch(100)     # so we will have 0 entities
    assert len(ll)==0


if __name__=="__main__":
    import os
    from google.appengine.api import apiproxy_stub_map
    from google.appengine.api import datastore_file_stub
    APP_ID="yours"
    os.environ['AUTH_DOMAIN'] = "yours.com"
    os.environ['APPLICATION_ID'] = APP_ID
    stub = datastore_file_stub.DatastoreFileStub(APP_ID, '/dev/null', '/dev/null')
    apiproxy_stub_map.apiproxy.RegisterStub('datastore_v3', stub)

    test_properties()
    test_selfreference()
    test_reference()
    test_delete_on_import()
    test_expando()
    test_poly()
    print "unittests ok"
